#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "ode-solver.h"

/*
 * This example solve system differential equations by two method:
 * Runge-Kutta and Runge-Kutta-Fehlberg.
 */


/*
 * dy[0] = y0[1];
 * dy[1] = 2.0 * y0[1] - y0[0];
 * this example equivalent to system:
 * y1' = y2
 * y2' = 2*y2 - y1
 * or equation
 * y'' - 2 * y' + y = 0
 * solution: y = t * exp(t)
 */

void
model(float t0, const float *y0, float *dy, void *opaque)
{
	assert(y0 != NULL && dy != NULL);

	dy[0] = y0[1];
	dy[1] = 2.0 * y0[1] - y0[0];
}

int
main(int argc, char *argv[])
{
	struct ode_solver *sol;
	struct ode_solver_system sys;
	float h, epsilon, t;
	float y0[2];
	float dy[2];
	float y[2];
	float y_err[2];
	int i, res;

	epsilon = 0.0001; /* tolerance                    */
	t       = 0.0;    /* initial time                 */
	h       = 1.0;    /* step for method              */
	y0[0]   = 0.0;    /* initial conditions y(0)  = 0 */
	y0[1]   = 1.0;    /* initial conditions y'(0) = 1 */
	y[0]    = 0.0;
	y[1]    = 0.0;

	sol           = ode_solver_rkf45_new();
	sys.model     = model; /* differential equations model */
	sys.opaque    = NULL;
	sys.dimension = 2;     /* differential equations order */

	/* find solution by t = 1 */
	while (t + h <= 1.0) {
		sol->solver_step(sol, &sys, t, h, y, y0, dy, y_err);

		res = ode_solver_modify_h(&sys, epsilon, &h, y_err);

		if (res == ODE_SOLVER_H_ACCEPT) {

			t = t + h;

			memcpy(y0, y, sys.dimension * sizeof(*y));
		}
	}

	/* last step correction*/
	if (t + h > 1.0 && t < 1.0) {
		h = 1.0 - t;

		sol->solver_step(sol, &sys, t, h, y, y0, dy, y_err);

		t += h;
	}

	printf("Method Runge-Kutta-Fehlberg\n");
	printf("t = %e y[0] = %e y[1] = %e\n",t, y[0], y[1]);

	ode_solver_free(sol);

	t     = 0.0;  /* initial time                 */
	h     = 0.05; /* step for method              */
	y0[0] = 0.0;  /* initial conditions y(0)  = 0 */
	y0[1] = 1.0;  /* initial conditions y'(0) = 1 */
	y[0]  = 0.0;
	y[1]  = 0.0;

	sol = ode_solver_rk4_new();

	/* find solution by t = 1 */
	for (i = 0; i < 20; i++) {

		sol->solver_step(sol, &sys, t, h, y, y0, dy, y_err);
		ode_solver_step(sol, &sys, t, h, y, y0, dy, y_err);

		memcpy(y0, y, sys.dimension * sizeof(*y));

		t += h;
	}

	ode_solver_free(sol);

	printf("Method Runge-Kutta\n");
	printf("t = %e y[0] = %e y[1] = %e\n",t, y[0], y[1]);

	return 0;
}
